package ru.dbotthepony.mc.otm.block.entity.decorative

import net.minecraft.core.BlockPos
import net.minecraft.server.level.ServerLevel
import net.minecraft.world.entity.item.ItemEntity
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.item.ItemStack
import net.minecraft.world.level.block.state.BlockState
import net.neoforged.neoforge.capabilities.Capabilities
import net.neoforged.neoforge.fluids.FluidStack
import net.neoforged.neoforge.fluids.capability.IFluidHandler
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kommons.util.ListenableDelegate
import ru.dbotthepony.mc.otm.block.entity.MatteryDeviceBlockEntity
import ru.dbotthepony.mc.otm.capability.item.CombinedItemHandler
import ru.dbotthepony.mc.otm.capability.fluid.BlockMatteryFluidHandler
import ru.dbotthepony.mc.otm.capability.moveFluid
import ru.dbotthepony.mc.otm.config.ItemsConfig
import ru.dbotthepony.mc.otm.container.HandlerFilter
import ru.dbotthepony.mc.otm.container.get
import ru.dbotthepony.mc.otm.container.slotted.AutomationFilter
import ru.dbotthepony.mc.otm.container.slotted.AutomationFilters
import ru.dbotthepony.mc.otm.container.slotted.ContainerSlot
import ru.dbotthepony.mc.otm.container.slotted.FilteredContainerSlot
import ru.dbotthepony.mc.otm.container.slotted.SlottedContainer
import ru.dbotthepony.mc.otm.core.isNotEmpty
import ru.dbotthepony.mc.otm.core.isNotSameAs
import ru.dbotthepony.mc.otm.menu.decorative.FluidTankMenu
import ru.dbotthepony.mc.otm.network.wrap
import ru.dbotthepony.mc.otm.registry.game.MBlockEntities

class FluidTankBlockEntity(blockPos: BlockPos, blockState: BlockState) : MatteryDeviceBlockEntity(MBlockEntities.FLUID_TANK, blockPos, blockState) {
	val fluid = BlockMatteryFluidHandler(ItemsConfig::FLUID_TANK_CAPACITY, syncher.add(ListenableDelegate.SmartBox(FluidStack.EMPTY, setter = { access, value ->
		access.accept(value)
		level?.lightEngine?.checkBlock(blockPos)

		if (level is ServerLevel) {
			markDirtyFast()
		}
	}), FluidStack.OPTIONAL_STREAM_CODEC.wrap()))

	private inner class FillSlot(container: SlottedContainer, slot: Int) : FilteredContainerSlot(container, slot) {
		override fun canAutomationPlaceItem(itemStack: ItemStack): Boolean {
			if (!super.canAutomationPlaceItem(itemStack))
				return false

			if (fluid.isEmpty) {
				return itemStack.getCapability(Capabilities.FluidHandler.ITEM)?.let { it.tanks > 0 } ?: false
			}

			return itemStack.getCapability(Capabilities.FluidHandler.ITEM)?.let { it.fill(fluid[0], IFluidHandler.FluidAction.SIMULATE) > 0 } ?: false
		}

		override fun canAutomationTakeItem(desired: Int): Boolean {
			return super.canAutomationTakeItem(desired) && !canAutomationPlaceItem(item)
		}
	}

	val inputContainer = SlottedContainer.Builder()
		.add(DRAIN_TAG, AutomationFilters.DRAINABLE_FLUID_CONTAINERS.filteredProvider)
		.add(FILL_TAG, ::FillSlot)
		.onChanged(::markDirtyFast)
		.build()
		.also(::addDroppableContainer)

	val outputContainer = SlottedContainer.Builder()
		.add(AutomationFilters.ONLY_OUT.simpleProvider)
		.onChanged(::markDirtyFast)
		.build()
		.also(::addDroppableContainer)

	private val fillSlot = inputContainer[FILL_TAG]
	private val drainSlot = inputContainer[DRAIN_TAG]

	val itemConfig = ConfigurableItemHandler(
		input = inputContainer,
		output = outputContainer,
		frontDefault = ItemHandlerMode.INPUT_OUTPUT,
		backDefault = ItemHandlerMode.INPUT_OUTPUT,
		leftDefault = ItemHandlerMode.INPUT_OUTPUT,
		rightDefault = ItemHandlerMode.INPUT_OUTPUT,
		topDefault = ItemHandlerMode.INPUT_OUTPUT,
		bottomDefault = ItemHandlerMode.INPUT_OUTPUT,
	)

	val fluidConfig = ConfigurableFluidHandler(fluid)

	init {
		savetables.stateful(::fluid, FLUID_KEY)
		savetables.stateful(::inputContainer)
		savetables.stateful(::outputContainer)
	}

	private fun drainItem() {
		val item = drainSlot.item

		if (item.isNotEmpty) {
			val cap = (if (item.count == 1) item else item.copyWithCount(1)).getCapability(Capabilities.FluidHandler.ITEM)

			if (cap == null) {
				if (outputContainer.consumeItem(item, simulate = false)) {
					drainSlot.setChanged()
				}

				return
			}

			if (fluid.isNotFull) {
				if (item.count == 1) {
					val moved0 = moveFluid(source = cap, destination = fluid)

					if (moved0.isNotEmpty) {
						drainSlot.item = cap.container

						if (outputContainer.consumeItem(drainSlot.item, simulate = false)) {
							drainSlot.setChanged()
						}
					}
				} else {
					val moved0 = moveFluid(source = cap, destination = fluid, actuallyFill = false)

					if (moved0.isNotEmpty) {
						if (outputContainer.consumeItem(cap.container, simulate = true)) {
							val cap1 = item.copyWithCount(1).getCapability(Capabilities.FluidHandler.ITEM) ?: throw ConcurrentModificationException()

							val moved1 = moveFluid(source = cap1, destination = fluid)

							if (moved1 isNotSameAs moved0) {
								LOGGER.error("Error moving fluids in Fluid tank at $blockPos: moved $moved0 during simulation from ${cap.container}, moved $moved1 from ${cap1.container} during execution. This is likely a bug in OTM or other mod!")
							} else {
								item.count--
								drainSlot.setChanged()

								if (!outputContainer.consumeItem(cap1.container, simulate = false)) {
									LOGGER.error("Unable to insert ${cap1.container} into output slot of Fluid tank at $blockPos, popping item in world instead to avoid item loss! This is likely a bug in OTM or other mod!")
									(level as? ServerLevel)?.addFreshEntity(ItemEntity(level!!, blockPos.x.toDouble(), blockPos.y.toDouble(), blockPos.z.toDouble(), cap1.container))
								}
							}
						}
					}
				}
			}
		}
	}

	private fun fillItem() {
		val item = fillSlot.item

		if (item.isNotEmpty) {
			val cap = (if (item.count == 1) item else item.copyWithCount(1)).getCapability(Capabilities.FluidHandler.ITEM)

			if (cap == null) {
				if (outputContainer.consumeItem(item, simulate = false)) {
					fillSlot.setChanged()
				}

				return
			}

			if (fluid.isNotEmpty) {
				if (item.count == 1) {
					val moved0 = moveFluid(source = fluid, destination = cap)

					if (moved0.isNotEmpty) {
						fillSlot.item = cap.container

						if (outputContainer.consumeItem(fillSlot.item, simulate = false)) {
							fillSlot.setChanged()
						}
					}
				} else {
					val moved0 = moveFluid(source = fluid, destination = cap, actuallyDrain = false)

					if (moved0.isNotEmpty) {
						if (outputContainer.consumeItem(cap.container, simulate = true)) {
							val cap1 = item.copyWithCount(1).getCapability(Capabilities.FluidHandler.ITEM) ?: throw ConcurrentModificationException()

							val moved1 = moveFluid(source = fluid, destination = cap1)

							if (moved1 isNotSameAs moved0) {
								LOGGER.error("Error moving fluids in Fluid tank at $blockPos: moved $moved0 during simulation from ${cap.container}, moved $moved1 from ${cap1.container} during execution. This is likely a bug in OTM or other mod!")
							} else {
								item.count--
								fillSlot.setChanged()

								if (!outputContainer.consumeItem(cap1.container, simulate = false)) {
									LOGGER.error("Unable to insert ${cap1.container} into output slot of Fluid tank at $blockPos, popping item in world instead to avoid item loss! This is likely a bug in OTM or other mod!")
									(level as? ServerLevel)?.addFreshEntity(ItemEntity(level!!, blockPos.x.toDouble(), blockPos.y.toDouble(), blockPos.z.toDouble(), cap1.container))
								}
							}
						}
					}
				}
			}
		}
	}

	override fun tick() {
		super.tick()

		drainItem()
		fillItem()
	}

	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu {
		return FluidTankMenu(containerID, inventory, this)
	}

	companion object {
		const val FLUID_KEY = "fluid"
		private val LOGGER = LogManager.getLogger()

		private val FILL_TAG = SlottedContainer.tag<FilteredContainerSlot>()
		private val DRAIN_TAG = SlottedContainer.tag<FilteredContainerSlot>()
	}
}
